#ifndef __MAILSYNC_UTILS__
#include <stdio.h>
#include <string>

using namespace std;

void print_with_escapes( FILE* f, const string& str);
char* nccs( const string& s);

#define __MAILSYNC_UTILS__
#endif
